#ifndef ZAD2_BASICBLOCK_H
#define ZAD2_BASICBLOCK_H
#include "../TypeDefs.h"
#include <utility>
#include <vector>
#include <map>
#include <cassert>
#include "Variable.h"
#include "Quadruple.h"
#include "Util.h"

using namespace std;

class BasicBlock : public std::enable_shared_from_this<BasicBlock> {
public:
	BasicBlock() : std::enable_shared_from_this<BasicBlock>() {
		phi = make_shared<QPhi>();
		flow.b = this;
		quads.emplace_back(phi);
	};
	vector<QuadruplePtr> quads;
	vector<QuadruplePtr> afterInit;
	vector<BasicBlockPtr> in, out;
	shared_ptr<QPhi> phi;
	void append(const BasicBlockPtr& after) {
		assert(after);
		out.push_back(after);
		after->in.push_back(shared_from_this());
		after->phi->phi.emplace_back();
	}

	std::string getName() const {
		for (const auto& q : quads) {
			if (auto l = dynamic_pointer_cast<QLabel>(q)) {
				return l->label;
			}
		}
		return "?";
	}

	void addJumpInitQuad(const QuadruplePtr& q) {
		afterInit.push_back(q);
	}

	map<VariablePtr, VariablePtr> &findPhi(const BasicBlockPtr& blk) {
		auto it = find(in.begin(), in.end(), blk);
		if (it == in.end()) throw runtime_error("blk not found in incoming blocks");
		auto mit = phi->phi.begin() + (it - in.begin());
		return *mit;
	}

	void addPhi(const BasicBlockPtr& blk, const VariablePtr& local, VariablePtr remote) {
		assert(blk && local && remote);
		assert(!local->info || !local->info->isInstanceVariable());
		auto &phiMap = findPhi(blk);
		phiMap[local] = std::move(remote);
	}

	void finishQuads();

	struct FlowAnalysisData {
		set<VariablePtr> in, out, def, use;
		BasicBlock *b {};
		void addUse(const VariablePtr& v) {
			if (!def.count(v)) use.emplace(v);
		}

		void addDef(const VariablePtr &v) {
			def.emplace(v);
			use.erase(v);
		}

		[[nodiscard]] set<VariablePtr> inForBlock(const BasicBlockPtr& q) const {
			set<VariablePtr> ret;
			for (const auto& var : b->findPhi(q)) {
				if (var.second->constExpr) continue;
				ret.emplace(var.second);
			}
			return in + ret;
		}
	};
	FlowAnalysisData flow;

	const map<VariablePtr, VariablePtr>& getPhiMapping(const BasicBlockPtr &b) const;
};


#endif //ZAD2_BASICBLOCK_H
